from importlib.metadata import version


def write_header(path):
    """Write the header for the new Zacros input file.

    Parameters
    ----------
    path: str
        Path of the input file
    """
    with open(path, 'w') as infile:
        infile.write('############################################################################\n')
        infile.write(f"#  Zacros input file generated by ZacrosTools (version {version('zacrostools')+')':<18}  #\n")
        infile.write('#  https://github.com/hprats/ZacrosTools                                   #\n')
        infile.write('#                                                                          #\n')
        infile.write('#  Hector Prats, PhD                                                       #\n')
        infile.write('############################################################################\n\n')



import numpy as np
from zacrostools.calc_functions import find_nearest


def parse_general_output(path):
    dmatch = {
        'n_gas_species': 'Number of gas species:',
        'gas_species_names': 'Gas species names:',
        'n_surf_species': 'Number of surface species:',
        'surf_species_names': 'Surface species names:',
        'n_sites': 'Total number of lattice sites:',
        'area': 'Lattice surface area:',
        'site_types': 'Site type names and total number of sites of that type:'
    }
    data = {}
    num_matches = 0
    with open(f"{path}/general_output.txt", 'r') as file_object:
        line = file_object.readline()
        while num_matches < len(dmatch):
            for key, pattern in dmatch.items():
                if pattern in line:
                    if key in ['n_gas_species', 'n_surf_species', 'n_sites']:
                        data[key] = int(line.split()[-1])
                    elif key == 'gas_species_names':
                        data[key] = line.split(':')[-1].split()
                    elif key == 'surf_species_names':
                        data[key] = [ads[0:-1] for ads in line.split(':')[-1].split()]
                    elif key == 'area':
                        data[key] = float(line.split()[-1])
                    elif key == 'site_types':
                        line = file_object.readline()
                        site_types = {}
                        while line.strip():
                            num_sites_of_given_type = int(line.strip().split(' ')[1].replace('(', '').replace(')', ''))
                            site_types[line.strip().split(' ')[0]] = num_sites_of_given_type
                            line = file_object.readline()
                        data[key] = site_types
                    num_matches += 1
            line = file_object.readline()
        return data


def parse_simulation_input(path):
    dmatch = ['pressure',
              'gas_specs_names',
              'gas_molar_fracs']
    data = {}
    with open(f"{path}/simulation_input.dat", 'r') as file_object:
        line = file_object.readline()
        while len(dmatch) != 0:
            if 'pressure' in line:
                data['pressure'] = float(line.split()[-1])
                dmatch.remove('pressure')
            elif 'gas_specs_names' in line:
                data['gas_specs_names'] = line.split()[1:]
                dmatch.remove('gas_specs_names')
            elif 'gas_molar_fracs' in line:
                data['gas_molar_fracs'] = [float(x) for x in line.split()[1:]]
                dmatch.remove('gas_molar_fracs')
            line = file_object.readline()
        return data


def get_data_specnum(path, ignore=0.0):
    with open(f"{path}/specnum_output.txt", "r") as infile:
        header = infile.readline().split()
    full_data = np.loadtxt(f"{path}/specnum_output.txt", skiprows=1)
    index = np.where(full_data[:, 2] == find_nearest(full_data[:, 2], float(full_data[-1, 2] * ignore / 100)))[0][0]
    data = np.delete(full_data, slice(0, index), 0)
    return data, header


def get_step_names(path):
    step_names = []
    with open(f"{path}/mechanism_input.dat", 'r') as file_object:
        line = file_object.readline()
        while 'end_mechanism' not in line:
            if 'reversible_step' in line and 'end_reversible_step' not in line:
                step_names.append(line.split()[-1])
            line = file_object.readline()
    return step_names



import os
import ast
import pandas as pd
from random import randint

import zacrostools.lattice_input
from zacrostools.write_functions import write_header
from zacrostools.mechanism_input import ReactionModel
from zacrostools.energetics_input import EnergeticModel
from zacrostools.custom_exceptions import *


class KMCModel:
    """A class that represents a KMC model.

    Parameters
    ----------
    gas_data: pd.DataFrame
        A Pandas DataFrame containing information about the gas molecules.
    mechanism_data: pd.DataFrame
        A Pandas DataFrame containing information about the reaction model.
    energetics_data: pd.DataFrame
        A Pandas DataFrame containing information about the energetic model.
    lattice_model: zacrostools.lattice_input.LatticeModel
        A lattice model
    """

    @enforce_types
    def __init__(self, gas_data: pd.DataFrame, mechanism_data: pd.DataFrame, energetics_data: pd.DataFrame,
                 lattice_model: zacrostools.lattice_input.LatticeModel):
        self.path = None
        self.gas_data = gas_data
        self.reaction_model = ReactionModel(mechanism_data=mechanism_data)
        self.energetic_model = EnergeticModel(energetics_data=energetics_data)
        self.lattice_model = lattice_model

        self.check_errors()

    def check_errors(self):
        """Checks for data consistency after initialization."""
        if self.lattice_model.lattice_type != 'default':
            if 'site_types' not in self.reaction_model.df.columns:
                raise ReactionModelError("Custom lattice used but no 'site_types' specified in the reaction model.")
            if 'site_types' not in self.energetic_model.df.columns:
                raise EnergeticModelError("Custom lattice used but no 'site_types' specified in the energetic model.")

    @enforce_types
    def create_job_dir(self,
                       # Mandatory arguments
                       path: str, temperature: Union[float, int], pressure: dict,
                       # Optional arguments
                       reporting_scheme: Union[dict, None] = None, stopping_criteria: Union[dict, None] = None,
                       manual_scaling: Union[dict, None] = None, auto_scaling_steps: Union[list, None] = None,
                       auto_scaling_tags: Union[dict, None] = None, sig_figs_energies: int = 16, sig_figs_pe: int = 16,
                       random_seed: Union[int, None] = None):
        """

        Parameters
        ----------
        path: str
            The path for the job directory where input files will be written
        temperature: float
            Reaction temperature (in K)
        pressure: dict
            Partial pressures of all gas species (in bar), e.g. {'CO': 1.0, 'O2': 0.001}
        reporting_scheme: dict, optional
            Reporting scheme in Zacros format. Must contain the following keys: 'snapshots', 'process_statistics' and
            'species_numbers'
            Default value: {'snapshots': 'on event 10000', 'process_statistics': 'on event 10000',
            'species_numbers': 'on event 10000'}
        stopping_criteria: dict, optional
            Stopping criteria in Zacros format. Must contain the following keys: 'max_steps', 'max_time' and 'wall_time'
            Default value: {'max_steps': 'infinity', 'max_time': 'infinity', 'wall_time': 86400}
        manual_scaling: dict, optional
            Step names (keys) and their corresponding manual scaling factors (values) e.g. {'CO_diffusion': 1.0e-1,
            'O_diffusion': 1.0e-2}
            Default value: {}
        auto_scaling_steps: list of str, optional
            Steps that will be marked as 'stiffness_scalable' in mechanism_input.dat.
            Default value: []
        auto_scaling_tags: dict, optional
            Keywords controlling the dynamic scaling algorithm and their corresponding values, e.g. {'check_every': 500,
            'min_separation': 400.0, 'max_separation': 600.0}.
            Default value: {}
        sig_figs_energies: int, optional
            Number of significant figures used when writing 'cluster_eng' in energetics_input.dat and 'activ_eng' in
            mechanism_input.dat.
            Default value: 16
        sig_figs_pe: int, optional
            Number of significant figures used when writing 'pre_expon' and 'pe_ratio' in mechanism_input.dat.
            Default value: 16
        random_seed: int, optional
            The integer seed of the random number generator. If not specified, ZacrosTools will generate one.
            Default value: None
        """

        if reporting_scheme is None:
            reporting_scheme = {'snapshots': 'on event 10000', 'process_statistics': 'on event 10000',
                                'species_numbers': 'on event 10000'}
        if stopping_criteria is None:
            stopping_criteria = {'max_steps': 'infinity', 'max_time': 'infinity', 'wall_time': 86400}
        if manual_scaling is None:
            manual_scaling = {}
        if auto_scaling_steps is None:
            auto_scaling_steps = []
        if auto_scaling_tags is None:
            auto_scaling_tags = {}
        if len(auto_scaling_steps) == 0 and len(auto_scaling_tags) > 0:
            raise InconsistentDataError("'auto_scaling_tags' defined but no steps are stiffness scalable")
        self.path = path
        if not os.path.exists(self.path):
            os.mkdir(self.path)
            self.write_simulation_input(temperature=temperature, pressure=pressure, reporting_scheme=reporting_scheme,
                                        stopping_criteria=stopping_criteria, auto_scaling_tags=auto_scaling_tags,
                                        sig_figs_energies=sig_figs_energies, random_seed=random_seed)
            self.reaction_model.write_mechanism_input(path=self.path, temperature=temperature, gas_data=self.gas_data,
                                                      manual_scaling=manual_scaling,
                                                      auto_scaling_steps=auto_scaling_steps,
                                                      sig_figs_energies=sig_figs_energies,
                                                      sig_figs_pe=sig_figs_pe)
            self.energetic_model.write_energetics_input(path=self.path,
                                                        sig_figs_energies=sig_figs_energies)
            self.lattice_model.write_lattice_input(path=self.path)
        else:
            print(f'{self.path} already exists (nothing done)')

    def write_simulation_input(self, temperature, pressure, reporting_scheme, stopping_criteria, auto_scaling_tags,
                               sig_figs_energies, random_seed):
        """Writes the simulation_input.dat file"""
        gas_specs_names = [x for x in self.gas_data.index]
        surf_specs = self.get_surf_specs()
        write_header(f"{self.path}/simulation_input.dat")
        with open(f"{self.path}/simulation_input.dat", 'a') as infile:
            if random_seed is None:
                infile.write('random_seed\t'.expandtabs(26) + str(randint(100000, 999999)) + '\n')
            else:
                infile.write('random_seed\t'.expandtabs(26) + str(random_seed) + '\n')
            infile.write('temperature\t'.expandtabs(26) + str(float(temperature)) + '\n')
            p_tot = sum(pressure.values())
            infile.write('pressure\t'.expandtabs(26) + str(p_tot) + '\n')
            infile.write('n_gas_species\t'.expandtabs(26) + str(len(gas_specs_names)) + '\n')
            infile.write('gas_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in gas_specs_names) + '\n')
            tags_dict = ['gas_energy', 'gas_molec_weight']
            tags_zacros = ['gas_energies', 'gas_molec_weights']
            for tag1, tag2 in zip(tags_dict, tags_zacros):
                tag_list = [self.gas_data.loc[x, tag1] for x in gas_specs_names]
                if tag1 == 'gas_energy':
                    formatted_tag_list = [f'{x:.{sig_figs_energies}f}' for x in tag_list]
                    infile.write(f'{tag2}\t'.expandtabs(26) + " ".join(formatted_tag_list) + '\n')
                else:
                    infile.write(f'{tag2}\t'.expandtabs(26) + " ".join(str(x) for x in tag_list) + '\n')
            try:
                gas_molar_frac_list = [pressure[x] / p_tot for x in gas_specs_names]
            except KeyError as ke:
                print(f"Key not found in 'pressure' dictionary: {ke}")
                print(f"When calling KMCModel.create_job_dir(), 'pressure' dictionary must contain the names of all "
                      f"gas species ")
            infile.write(f'gas_molar_fracs\t'.expandtabs(26) + " ".join(str(x) for x in gas_molar_frac_list) + '\n')
            infile.write('n_surf_species\t'.expandtabs(26) + str(len(surf_specs)) + '\n')
            infile.write('surf_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs.keys()) + '\n')
            infile.write('surf_specs_dent\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs.values()) + '\n')
            for tag in ['snapshots', 'process_statistics', 'species_numbers']:
                infile.write((tag + '\t').expandtabs(26) + str(reporting_scheme[tag]) + '\n')
            for tag in ['max_steps', 'max_time', 'wall_time']:
                infile.write((tag + '\t').expandtabs(26) + str(stopping_criteria[tag]) + '\n')
            if len(auto_scaling_tags) > 0:
                infile.write(f"enable_stiffness_scaling\n")
                for tag in auto_scaling_tags:
                    infile.write((tag + '\t').expandtabs(26) + str(auto_scaling_tags[tag]) + '\n')
            infile.write(f"finish\n")

    def get_surf_specs(self):
        # Identify all surf_specs and their corresponding dentates from the energetic_model dataframe
        # Used to write 'surf_specs_names' and 'surf_specs_dent' in the simulation_input.dat file
        surf_specs = {}
        for cluster in self.energetic_model.df.index:
            lattice_state = ast.literal_eval(self.energetic_model.df.loc[cluster, 'lattice_state'])
            for site in lattice_state:
                surf_specs_name = site.split()[1]
                surf_specs_dent = int(site.split()[2])
                if surf_specs_name not in surf_specs or (
                        surf_specs_name in surf_specs and surf_specs_dent > surf_specs[surf_specs_name]):
                    surf_specs[surf_specs_name] = surf_specs_dent
        return surf_specs



from zacrostools.write_functions import write_header
from zacrostools.custom_exceptions import LatticeModelError, enforce_types


class LatticeModel:
    """A class that represents a lattice model.

    Parameters
    ----------
    lines: list of str
        Lines that will be printed in the lattice_input.dat.
    """

    @enforce_types
    def __init__(self, lines: list = None, lattice_type: str = None):
        self.lines = lines
        self.lattice_type = lattice_type

    @classmethod
    @enforce_types
    def from_file(cls, path: str):
        """Create a LatticeModel by reading an existing lattice_input.dat file.

        Parameters
        ----------
        path: str
            Path to the directory where the lattice_input.dat file is located


        Returns
        -------
        lattice_model: LatticeModel

        """

        lattice_lines = []
        start_pattern_detected = False
        end_pattern_detected = False

        with open(path, 'r') as infile:

            while not start_pattern_detected:
                line = infile.readline()
                if 'default_choice' in line or 'periodic_cell' in line or 'explicit' in line:
                    lattice_lines.append(line)
                    start_pattern_detected = True

            if 'default_choice' in line:
                lattice_type = 'default'
            elif 'periodic_cell' in line:
                lattice_type = 'custom_periodic'
            else:
                lattice_type = 'custom_non_periodic'

            while not end_pattern_detected:
                line = infile.readline()
                if 'end_lattice' in line:
                    lattice_lines.append(line)
                    end_pattern_detected = True
                else:
                    lattice_lines.append(line)

        lattice_model = cls(lines=lattice_lines, lattice_type=lattice_type)
        return lattice_model

    @enforce_types
    def write_lattice_input(self, path: str):
        """Write the lattice_input.dat file.

        Parameters
        ----------
        path: str
            Path to the directory where the lattice_input.dat file will be written.

        """
        write_header(f"{path}/lattice_input.dat")
        with open(f"{path}/lattice_input.dat", 'a') as infile:
            for line in self.lines:
                infile.write(line)

    @enforce_types
    def repeat_cell(self, repeat_cell: list):
        """Modify the value of the repeat_cell keyword in the lattice_input.dat file.

        Parameters
        ----------
        repeat_cell: list of int, optional
            Updates the repeat_cell keyword in lattice_input.dat file.

        """
        if self.lattice_type == 'custom_non_periodic':
            raise LatticeModelError("repeat_cell()' method can not be used with custom non-periodic lattices.")
        i = 0
        line = self.lines[i]
        if self.lattice_type == 'custom_periodic':
            while 'repeat_cell' not in line:
                i += 1
                line = self.lines[i]
            self.lines[i] = f'   repeat_cell {repeat_cell[0]} {repeat_cell[1]}\n'
        else:
            while not any(keyword in line for keyword in
                          ['triangular_periodic', 'rectangular_periodic', 'hexagonal_periodic']):
                i += 1
                line = self.lines[i]
            keyword = line.split()[0]
            lattice_constant = line.split()[1]
            self.lines[i] = f'   {keyword} {lattice_constant} {repeat_cell[0]} {repeat_cell[1]}\n'
